@@ -1,4 +1,4 @@
-#$Id: Daemon.pm 761 2011-05-18 18:37:53Z fil $
+#$Id: Daemon.pm 911 2012-07-29 13:52:13Z monaco $
 ########################################################
 package POE::Component::Daemon;
 
@@ -11,11 +11,13 @@ use POSIX qw(EAGAIN ECHILD SIGINT SIGKILL SIGTERM);
 use POE;
 use Carp;
 use Data::Dumper;
+use File::Basename qw( basename );
+use File::Copy qw( move );
 use Scalar::Util qw( blessed );
 
 use POE::Component::Daemon::Scoreboard;
 
-$VERSION = '0.1300';
+$VERSION = '0.1301';
 
 sub DEBUG () { 0 }
 sub DEBUG_SC () { DEBUG or 0 }
@@ -27,7 +29,6 @@ BEGIN {
     # warn $NO_PEEK;
 }
 
-
 ########################################################
 sub new
 {
@@ -87,9 +88,10 @@ sub drop_privs
 sub open_logfile
 {
     my($self)=@_;
-    my $logfile = $self->{logfile};
+    my $logfile = $self->logfile_name;
     return unless $logfile;
     return if $self->{"logfile done"}++;
+    DEBUG and warn "Log file is $logfile\n";
     open STDOUT, ">>$logfile" or die "Unable to write to $logfile: $!\n";   
     open STDERR, ">&STDOUT"   or die "Unable to reopen STDERR: $!\n";
     STDERR->autoflush(1);
@@ -97,6 +99,36 @@ sub open_logfile
     return 1;
 }
 
+sub logfile_name
+{
+    my( $self ) = @_;
+    return unless $self->{logfile};
+    my $logfile = $self->{logfile};
+    if( $self->{logfile_per_process} and $self->{'is a child'} ) {
+        $logfile .= "-$$";
+    }
+    return $logfile;
+}
+
+
+sub reopen_logfile
+{
+    my( $self ) = @_;
+    delete $self->{"logfile done"};
+    $self->open_logfile;
+    DEBUG and warn "Start log\n";
+}
+
+sub rotate_logfile
+{
+    my( $self, $dir ) = @_;
+    my $logfile = $self->logfile_name;
+    return unless $logfile;
+    my $file = basename $logfile;
+    my $dest = File::Spec->catfile( $dir, $file );
+    eval { move $file, $dest } or warn $@;
+}
+
 ########################################################
 sub detach
 {
@@ -201,22 +233,22 @@ sub default_min_max
     my($self)=@_;
     if($self->{max_children}) {
         if($self->{max_spare}) {
-            $self->{min_spare} ||= int($self->{max_spare} /2);
+            $self->{min_spare} = int($self->{max_spare} /2) unless defined $self->{min_spare};
         }
         else {
-            $self->{min_spare} ||= int($self->{max_children} * 0.2);
-            $self->{max_spare} ||= int($self->{max_children} * 0.8);
+            $self->{min_spare} = int($self->{max_children} * 0.2) unless defined $self->{min_spare};
+            $self->{max_spare} = int($self->{max_children} * 0.8) unless defined $self->{max_spare};
         }
-        $self->{min_spare} ||=1;
-        $self->{max_spare} ||=2;
+        $self->{min_spare} = 1  unless defined $self->{min_spare};
+        $self->{max_spare} = 2  unless defined $self->{max_spare};
         if( $self->{max_spare} <  $self->{min_spare} ) {
             confess "Max_spare can't be smaller then $self->{min_spare}; madness follows.";
         }
     }
     else {
         # We couldn't be here unless start_children is set
-        $self->{min_spare} ||= $self->{start_children};
-        $self->{max_spare} ||= 2*$self->{min_spare};
+        $self->{min_spare} = $self->{start_children} unless defined $self->{min_spare};
+        $self->{max_spare} = 2*$self->{min_spare} unless defined $self->{max_spare};
         $self->{max_children} = $self->{start_children} 
                                     + $self->{max_spare};
     }
@@ -478,14 +510,18 @@ sub _stop
 # Someone wants us to exit... oblige
 sub do_shutdown
 {
-    # print STDERR "$$: do_shutdown\n";
+    DEBUG and print STDERR "$$: do_shutdown\n";
     $poe_kernel->call( $poe_kernel->get_active_session, 'shutdown' );
 }
+
 sub shutdown
 {
     my($self, $kernel)=@_[OBJECT, KERNEL];
 
-    $self->{verbose} and 
+    # $main::TRACE_REFCNT = 1;
+    # Daemon->peek(1);
+
+    # $self->{verbose} and 
         warn "$$: shutdown";
     if($self->{rogues}) {
         $kernel->delay('rogues');   # we no longer care about rogues
@@ -522,6 +558,7 @@ sub shutdown
     $kernel->sig( 'HUP'  );
     $kernel->sig( 'INT'  ); 
     $kernel->sig( 'TERM' ); 
+
     return;
 }
 
@@ -657,9 +694,16 @@ sub become_child
     # remove the wait for checking the scorebard
     $poe_kernel->delay('check_scoreboard') if $self->is_prefork or
                                               $self->is_fork;
+
+    foreach my $pid ( keys %{ $self->{children} || {} } ) {
+        $poe_kernel->sig_child( $pid );
+    }
+
     # remove these fields
     delete @{$self}{ qw(rogues proctable children) };
 
+    $self->reopen_logfile;
+
     # Tell everyone we are now a child
     $self->inform_others( 'daemon_child', $req );
 
@@ -768,6 +812,7 @@ sub sig_CHLD
     return if $self->{"is a child"};
 
 #    ( DEBUG or $self->{verbose} ) and 
+    $INC{'Test/More.pm'} or
         warn "$$: SIGCHLD pid=$pid";
 
     ##########
@@ -912,12 +957,9 @@ sub check_scoreboard
         DEBUG and warn "$$: child at slot $slot ($pid: $slots->[$slot])";
         if($slots->[$slot] eq 'w' or 
             $slots->[$slot] eq 'f') {      # waiting for req
-            warn "$pid is still forking" if $slots->[$slot] eq 'f';
+            warn "$pid is still forking\n" if DEBUG and $slots->[$slot] eq 'f';
             push @waiting, $pid;
         }
-        else {
-        }
-        
     }
 
 
@@ -926,17 +968,23 @@ sub check_scoreboard
         DEBUG and warn "$$: waiting=$waiting";
         if($waiting < $self->{min_spare}) {
             my $n=$self->{min_spare} - $waiting;
-            DEBUG_SC and 
+            # DEBUG_SC and 
                 warn "$$: Spawning $n spares";
             $self->fork_off($n);
         }
-        if($waiting > $self->{max_spare}) {
+        elsif($waiting > $self->{max_spare}) {
             my $n=$waiting - $self->{max_spare};
-            DEBUG_SC and warn "$$: Killing $n spares";
+            # DEBUG_SC and 
+                warn "$$: Killing $n spares";
             foreach my $pid ( @waiting[0..($n-1)] ) {
                 kill SIGINT, $pid or warn "$$: killing $pid: $!";
             }
         }
+        elsif( not $self->{statis} ) {
+            $INC{'Test/More.pm'} or
+                warn "All systems are go\n";
+            $self->{statis} = 1;
+        }
     }
     elsif( $self->is_fork and 
                     $self->{max_children} <= keys %{$self->{children}} ) {
@@ -998,16 +1046,18 @@ sub update_status_prefork
         warn "$$: current_status=$current_status -> status=$status";
 
     if($status eq 'wait' or $status eq 'done') {
-        DEBUG and warn "$$: Moving to status=wait";
+        # DEBUG and 
+            warn "$$: Moving to status=wait";
         $self->{scoreboard}->write($slot, 'wait');
         if($self->{requestN} >= $self->{'max requests'}) {
-            DEBUG and 
+            # DEBUG and 
                 warn "$$: Handled $self->{requestN} requests, shutting down status=$status";
             $self->do_shutdown;
         }
         elsif($current_status ne 'w') {
             # use Carp qw( cluck );
             # cluck "$$: daemon_accept";
+            warn "$$: requestN=$self->{requestN} < max requests=$self->{'max requests'}";
             $self->inform_others( 'daemon_accept' );
         }
         else {
@@ -1170,11 +1220,16 @@ sub peek
         $ret .= "\t\tEvent: ".$item->{event}."\n";
 
         if($verbose) {
-            $events->{ $item->{source}->ID }{source} ++; 
+            if( $item->{source}->ID == $item->{destination}->ID ) {
+                $events->{ $item->{destination}->ID }{source_destination} ++; 
+            }
+            else {
+                $events->{ $item->{source}->ID }{source} ++; 
+                $events->{ $item->{destination}->ID }{destination} ++; 
+            }
             $ret .= "\t\tSource: ".
                     $api->session_id_loggable($item->{source}).
                     "\n";
-            $events->{ $item->{destination}->ID }{destination} ++; 
             $ret .= "\t\tDestination: ".
                     $api->session_id_loggable($item->{destination}).
                     "\n";
@@ -1197,7 +1252,7 @@ sub peek
             $ref += $q1;
             $ret.="\t\textref count: $q1 (Stay alive)\n" if $q1;
 
-            $q1=$api->session_handle_count($session);
+            $q1=$api->session_handle_count($session->ID);
             $ref += $q1;  
             $ret.="\t\thandle count: $q1 (Stay alive)\n" if $q1;
 
@@ -1206,6 +1261,14 @@ sub peek
             $q1=join ',', @aliases;
             $ret.="\t\tAliases: $q1\n" if $q1;
 
+            if( $poe_kernel->can( '_data_sig_session_awaits_pids') ) {
+                $q1 = $poe_kernel->_data_sig_session_awaits_pids( $session->ID );
+            } else {
+                $q1 = $poe_kernel->_data_sig_pids_ses( $session->ID );
+            }
+            $ref += $q1;
+            $ret .= "\t\tPID count: $q1 (Stay alive)\n" if $q1;
+
             my @children = $api->get_session_children($session);
             if(@children) {
                 $ref += @children;
@@ -1225,6 +1288,12 @@ sub peek
                 $ref += $q1;
             }
 
+            $q1 = $events->{ $session->ID }{source_destination};
+            if( $q1 ) {
+                $ret.="\t\tEvent source+destination count: $q1 (Stay alive)\n";
+                $ref += $q1;
+            }
+
             if($refcount != $ref) {
                 $ret.="\t\tStay alive: refcount=$refcount counted=$ref\n";
             }  
@@ -1595,7 +1664,8 @@ Now, instead of cursing the $GODS because your kernel doesn't exit when you
 think it should, you simply type the following in another window.
 
     kill -USR2 I<pid>
-
+    # or, if you are on Solaris or Linux
+    pkill -USR2 -o your-daemon
 
 
 
@@ -1733,7 +1803,7 @@ Philip Gwyn, E<lt>gwyn -AT- cpan.orgE<gt>
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright 2004-2011 by Philip Gwyn. All rights reserved.
+Copyright 2004-2013 by Philip Gwyn. All rights reserved.
 
 This library is free software; you can redistribute it and/or modify
 it under the same terms as Perl itself. 
